"""
Modifiers
===========
When a request for configuration arrives at the control plane, it will first
retrieve data from all configured sources.

Directly after it has run all sources, it will apply modifications to the data.
If no modifiers are configured, the data is returned as-is.

Modifiers are enabled by adding the following configuration:

.. code-block:: yaml

   modifiers:
     - <modifier_name>

   global_modifiers:
     - <global_modifier_name>

The modifier name must match the name of the entry point

`todo add guide for adding entry points`

Modifiers will apply to each instance of the source data, whereas
global modifiers can apply to the entire set of source data.

Global modifiers are executed before modifiers.
"""
from typing import List
from pkg_resources import iter_entry_points
from sovereign import config
from sovereign.statistics import stats


_entry_points = iter_entry_points('sovereign.modifiers')
_modifiers = {ep.name: ep.load()
              for ep in _entry_points
              if ep.name in list(config.modifiers)}

_gentry_points = iter_entry_points('sovereign.global_modifiers')
_gmodifiers = {ep.name: ep.load()
               for ep in _gentry_points
               if ep.name in list(config.global_modifiers)}


@stats.timed('modifiers.apply_ms')
def apply_modifications(source_data: List[dict]) -> List[dict]:
    """
    Runs all configured modifiers on received data from sources.
    Returns the data, with modifications applied.
    """
    for g in _gmodifiers.values():
        global_modifier = g(source_data)
        global_modifier.apply()
        source_data = global_modifier.join()

    for index, instance in enumerate(source_data):
        for m in _modifiers.values():
            modifier = m(instance)
            if modifier.match():
                source_data[index] = modifier.apply()
    return source_data



import logging
from functools import wraps
from sovereign import config

try:
    from datadog import statsd as statsd
except ImportError:
    if config.statsd.enabled:
        raise
    statsd = None


class StatsDProxy:
    def __init__(self, statsd_instance=None):
        self.statsd = statsd_instance

    def __getattr__(self, item):
        if self.statsd is not None:
            return getattr(self.statsd, item)
        try:
            return StatsdNoop
        except TypeError:
            return self.do_nothing

    def do_nothing(self, *args, **kwargs):
        pass


class StatsdNoop:
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def __call__(self, func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapped


def configure_statsd(module):
    if config.statsd.enabled:
        module.host = config.statsd.host
        module.port = config.statsd.port
        module.namespace = config.statsd.namespace
        module.use_ms = config.statsd.use_ms
        for tag, value in config.statsd.loaded_tags.items():
            module.constant_tags.extend([f'{tag}:{value}'])
    else:
        module = None
        statsd_logger = logging.getLogger('datadog.dogstatsd')
        statsd_logger.disabled = True
    return StatsDProxy(module)


stats = configure_statsd(module=statsd)



"""
Sources are used to retrieve arbitrary data that can then be represented
in templates as envoy configuration.

A new custom source can be added by subclassing the Source class and
adding an entry point using setuptools.

`todo add guide for adding entry points`

When the control plane receives a request for configuration from an envoy
proxy, it first runs through all the configured sources and combines the
data received from them into a single list, to be used within templates.

The results are cached for a configurable number of seconds to allow several proxies to poll
at the same time, receiving data that is consistent with each other.
"""
import schedule
from glom import glom
from copy import deepcopy
from typing import List, Iterable
from pkg_resources import iter_entry_points
from sovereign import config
from sovereign.statistics import stats
from sovereign.schemas import DiscoveryRequest, Source
from sovereign.logs import LOG
from sovereign.modifiers import apply_modifications

_source_data = list()
_entry_points = iter_entry_points('sovereign.sources')
_sources = {e.name: e.load() for e in _entry_points}

if not _sources:
    raise RuntimeError('No sources available.')


def is_debug_request(v):
    return v == '' and config.debug_enabled


def is_wildcard(v):
    return v in [['*'], '*', ('*',)]


def contains(container, item):
    if isinstance(container, Iterable):
        return item in container


def setup_sources(source: Source):
    """
    Takes Sources from config and turns them into Python objects, ready to use.
    """
    cls = _sources[source.type]
    instance = cls(source.config)
    instance.setup()
    return instance


def pull_sources():
    """
    Runs .get() on every configured Source; returns the results in a generator.
    """
    for source in config.sources:
        instance = setup_sources(source)
        yield from instance.get()


def sources_refresh():
    """
    All source data is stored in ``sovereign.sources._source_data``.
    Since the variable is outside this functions scope, we can only make
    in-place modifications to it via its methods.

    This function retrieves all sources, puts them into a temporary list,
    and then clears and re-fills ``_source_data`` with the new data.

    The process is done in two steps to avoid ``_source_data`` being empty
    for any significant amount of time.
    """
    with stats.timed('sources.poll_time_ms'):
        new_sources = list()
        for source in pull_sources():
            if not isinstance(source, dict):
                LOG.msg('Received a non-dictionary source', level='warn', source_repr=repr(source))
                continue
            new_sources.append(source)

    if new_sources == _source_data:
        stats.increment('sources.unchanged')
        return
    else:
        stats.increment('sources.refreshed')

    with stats.timed('sources.swap_time_ms'):
        _source_data.clear()
        _source_data.extend(new_sources)
    return


def read_sources():
    """
    Returns a copy of source data in order to ensure it is not
    modified outside of the refresh function
    """
    return deepcopy(_source_data)


def match_node(request: DiscoveryRequest, modify=True) -> List[dict]:
    """
    Checks a node against all sources, using the node_match_key and source_match_key
    to determine if the node should receive the source in its configuration.

    :param request: envoy discovery request
    :param modify: switch to enable or disable modifications via Modifiers
    """
    ret = list()
    for source in read_sources():
        source_value = glom(source, config.source_match_key)
        node_value = glom(request.node, config.node_match_key)

        conditions = (
            is_debug_request(node_value),
            node_value == source_value,
            contains(source_value, node_value),
            is_wildcard(node_value),
            is_wildcard(source_value),
            config.node_matching is False
        )
        if any(conditions):
            ret.append(source)
    if modify:
        return apply_modifications(ret)
    return ret


def available_service_clusters():
    """
    Checks for all match keys present in existing sources and adds them to a list

    A dict is used instead of a set because dicts cannot have duplicate keys, and
    have ordering since python 3.6
    """
    ret = dict()
    ret['*'] = None
    for source in read_sources():
        source_value = glom(source, config.source_match_key)
        if isinstance(source_value, Iterable):
            for item in source_value:
                ret[item] = None
            continue
        ret[source_value] = None
    return list(ret.keys())


if __name__ != '__main__':
    schedule.every(config.sources_refresh_rate).seconds.do(sources_refresh)



import schedule
from fastapi import Body, BackgroundTasks
from fastapi.routing import APIRouter
from sovereign.logs import add_log_context
from starlette.responses import UJSONResponse
from sovereign import discovery, config
from sovereign.statistics import stats
from sovereign.schemas import DiscoveryRequest, DiscoveryResponse
from sovereign.utils.auth import authenticate

router = APIRouter()


@router.post(
    '/discovery:{xds_type}',
    summary='Envoy Discovery Service Endpoint',
    responses={
        200: {
            'model': DiscoveryResponse,
            'description': 'New resources provided'
        },
        config.no_changes_response_code: {
            'description': 'Resources are up-to-date'
        }
    }
)
async def discovery_response(
        xds_type: discovery.DiscoveryTypes,
        background_tasks: BackgroundTasks,
        discovery_request: DiscoveryRequest = Body(None)
):
    authenticate(discovery_request)
    response: dict = await discovery.response(discovery_request, xds_type.value)
    if response['version_info'] == discovery_request.version_info:
        ret = 'No changes'
        code = config.no_changes_response_code
    elif len(response.get('resources', [])) == 0:
        ret = 'No resources found'
        code = 404
    elif response['version_info'] != discovery_request.version_info:
        ret = response
        code = 200
    else:
        ret = 'Unknown Error'
        code = 500
    metrics_tags = [
        f"client_ip:{discovery_request.node.metadata.get('ipv4', '-')}",
        f"client_version:{discovery_request.envoy_version}",
        f"response_code:{code}",
        f"xds_type:{xds_type.value}"
    ]
    metrics_tags += [f"resource:{resource}" for resource in discovery_request.resource_names]
    stats.increment('discovery.rq_total', tags=metrics_tags)
    add_log_context(
        resource_names=discovery_request.resource_names,
        envoy_ver=discovery_request.envoy_version
    )
    background_tasks.add_task(schedule.run_pending)
    return UJSONResponse(content=ret, status_code=code)



"""
Discovery
---------

Functions used to render and return discovery responses to Envoy proxies.

The templates are configurable. `todo See ref:Configuration#Templates`
"""
import zlib
import yaml
from yaml.parser import ParserError
from enum import Enum
from jinja2 import meta
from starlette.exceptions import HTTPException
from sovereign import XDS_TEMPLATES
from sovereign.statistics import stats
from sovereign.context import template_context
from sovereign.sources import match_node
from sovereign.config_loader import jinja_env
from sovereign.schemas import XdsTemplate, DiscoveryRequest
from sovereign.utils.crypto import disabled_suite

try:
    default_templates = XDS_TEMPLATES['default']
except KeyError:
    raise KeyError(
        'Your configuration should contain default templates. For more details, see '
        'https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/tutorial.html#create-templates '
    )

# Create an enum that bases all the available discovery types off what has been configured
discovery_types = (_type for _type in sorted(XDS_TEMPLATES['__any__'].keys()))
DiscoveryTypes = Enum('DiscoveryTypes', {t: t for t in discovery_types})


@stats.timed('discovery.version_hash_ms')
def version_hash(*args) -> str:
    """
    Creates a 'version hash' to be used in envoy Discovery Responses.
    """
    data: bytes = repr(args).encode()
    version_info = zlib.adler32(data)
    return str(version_info)


def make_context(request: DiscoveryRequest, template: XdsTemplate):
    """
    Creates context variables to be passed into either a jinja template,
    or as kwargs to a python template.
    """
    context = {
        'instances': match_node(request),
        'resource_names': request.resources,
        **template_context
    }

    # If the discovery request came from a mock, it will
    # typically contain this metadata key.
    # This means we should prevent any decryptable data
    # from ending up in the response.
    if request.node.metadata.get('hide_private_keys'):
        context['crypto'] = disabled_suite

    if template.is_python_source:
        return context
    else:
        # Jinja templates will be converted to an AST and then scanned for unused
        # variables, which reduces computation in cases where a lot of context
        # has been generated, but does not need to be checksum'd or rendered.
        template_ast = jinja_env.parse(template.source)
        used_variables = meta.find_undeclared_variables(template_ast)
        for key in list(context):
            if key in used_variables:
                continue
            context.pop(key, None)
        return context


async def response(request: DiscoveryRequest, xds_type):
    """
    A Discovery **Request** typically looks something like:

    .. code-block:: json

        {
            "version_info": "0",
            "node": {
                "cluster": "T1",
                "build_version": "<revision hash>/<version>/Clean/RELEASE",
                "metadata": {
                    "auth": "..."
                }
            }
        }

    When we receive this, we give the client the latest configuration via a
    Discovery **Response** that looks something like this:

    .. code-block:: json

        {
            "version_info": "abcdef1234567890",
            "resources": []
        }

    The version_info is derived from :func:`sovereign.discovery.version_hash`

    :param request: An envoy Discovery Request
    :param xds_type: what type of XDS template to use when rendering
    :return: An envoy Discovery Response
    """
    metrics_tags = [
        f'xds_type:{xds_type}',
        f'partition:{request.node.cluster}'
    ]
    with stats.timed('discovery.total_ms', tags=metrics_tags):
        template: XdsTemplate = XDS_TEMPLATES.get(request.envoy_version, default_templates)[xds_type]
        context = make_context(request, template)

        config_version = version_hash(context, template.checksum, request.node.common)
        if config_version == request.version_info:
            return {'version_info': config_version}

        if template.is_python_source:
            envoy_configuration = {
                'resources': list(template.code.call(discovery_request=request, **context)),
                'version_info': config_version
            }
        else:
            with stats.timed('discovery.render_ms', tags=metrics_tags):
                rendered = await template.content.render_async(discovery_request=request, **context)
            with stats.timed('discovery.deserialize_ms', tags=metrics_tags):
                try:
                    envoy_configuration = yaml.safe_load(rendered)
                    envoy_configuration['version_info'] = config_version
                except ParserError:
                    raise HTTPException(
                        status_code=500,
                        detail='Failed to load configuration, there may be '
                               'a syntax error in the configured templates.'
                    )
        return remove_unwanted_resources(envoy_configuration, request.resources)


def remove_unwanted_resources(conf, requested):
    """
    If Envoy specifically requested a resource, this removes everything
    that does not match the name of the resource.
    If Envoy did not specifically request anything, every resource is retained.
    """
    ret = dict()
    ret['version_info'] = conf['version_info']
    ret['resources'] = [
        resource
        for resource in conf['resources']
        if resource_name(resource) in requested
    ]
    return ret


def resource_name(resource):
    return resource.get('name') or resource['cluster_name']



from collections import defaultdict
from fastapi import APIRouter, Query, Path, Cookie
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse
from sovereign import html_templates, discovery, XDS_TEMPLATES
from sovereign.discovery import DiscoveryTypes
from sovereign.sources import available_service_clusters
from sovereign.utils.mock import mock_discovery_request

router = APIRouter()

all_types = [t.value for t in DiscoveryTypes]


@router.get('/')
async def ui_main():
    return RedirectResponse(url=f'/ui/resources/{all_types[0]}')


@router.get('/set-version')
async def set_envoy_version(
    request: Request,
    version: str = Query('__any__', title='The clients envoy version to emulate in this XDS request'),
):
    url = request.headers.get('Referer', '/ui')
    response = RedirectResponse(url=url)
    response.set_cookie(key='envoy_version', value=version)
    return response


@router.get('/set-service-cluster')
async def set_service_cluster(
    request: Request,
    service_cluster: str = Query('__any__', title='The clients envoy version to emulate in this XDS request'),
):
    url = request.headers.get('Referer', '/ui')
    response = RedirectResponse(url=url)
    response.set_cookie(key='service_cluster', value=service_cluster)
    return response


@router.get('/resources/{xds_type}')
async def resources(
        request: Request,
        xds_type: DiscoveryTypes = Path('clusters', title='xDS type', description='The type of request'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        service_cluster: str = Cookie('*', title='The clients service cluster to emulate in this XDS request'),
        envoy_version: str = Cookie('__any__', title='The clients envoy version to emulate in this XDS request')
):
    ret = defaultdict(list)
    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=[],
        version=envoy_version,
        region=region
    )
    try:
        response = await discovery.response(
            request=mock_request,
            xds_type=xds_type.value
        )
    except KeyError:
        ret['resources'] = []
    else:
        if isinstance(response, dict):
            ret['resources'] += response.get('resources') or []
    return html_templates.TemplateResponse(
        name='resources.html',
        media_type='text/html',
        context={
            'resources': ret['resources'],
            'request': request,
            'resource_type': xds_type.value,
            'all_types': all_types,
            'version': envoy_version,
            'available_versions': list(XDS_TEMPLATES.keys()),
            'service_cluster': service_cluster,
            'available_service_clusters': available_service_clusters(),
        })


@router.get('/resources/{xds_type}/{resource_name}')
async def resource(
        xds_type: DiscoveryTypes = Path('clusters', title='xDS type', description='The type of request'),
        resource_name: str = Path(..., title='Name of the resource to view'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        service_cluster: str = Cookie('*', title='The clients service cluster to emulate in this XDS request'),
        envoy_version: str = Cookie('__any__', title='The clients envoy version to emulate in this XDS request')
):
    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=[],
        version=envoy_version,
        region=region
    )
    response = await discovery.response(
        request=mock_request,
        xds_type=xds_type.value
    )
    if isinstance(response, dict):
        for res in response.get('resources', []):
            name = res.get('name') or res['cluster_name']
            if name == resource_name:
                return JSONResponse(content=res)


@router.get('/resources/routes/{route_configuration}/{virtual_host}')
async def virtual_hosts(
        route_configuration: str = Path(..., title='Name of the route configuration'),
        virtual_host: str = Path(..., title='Name of the resource to view'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        service_cluster: str = Cookie('*', title='The clients service cluster to emulate in this XDS request'),
        envoy_version: str = Cookie('__any__', title='The clients envoy version to emulate in this XDS request')
):
    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=[],
        version=envoy_version,
        region=region
    )
    response = await discovery.response(
        request=mock_request,
        xds_type='routes'
    )
    if isinstance(response, dict):
        route_config = [
            r for r in response.get('resources', [])
            if r['name'] == route_configuration
        ][0]

        for vhost in route_config['virtual_hosts']:
            if vhost['name'] == virtual_host:
                return JSONResponse(content=vhost)
